/*
 * Copyright (C) 2021 Open Source Robotics Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
*/
#ifndef GZ_SENSORS_WIDEANGLECAMERASENSOR_HH_
#define GZ_SENSORS_WIDEANGLECAMERASENSOR_HH_

#include <memory>
#include <cstdint>
#include <string>

#include <sdf/sdf.hh>

#include <gz/common/Event.hh>
#include <gz/utils/SuppressWarning.hh>

#include <gz/msgs/image.pb.h>

// TODO(louise) Remove these pragmas once gz-rendering is disabling the
// warnings
#ifdef _WIN32
#pragma warning(push)
#pragma warning(disable: 4251)
#endif
#include <gz/rendering/WideAngleCamera.hh>
#ifdef _WIN32
#pragma warning(pop)
#endif

#include "gz/sensors/wide_angle_camera/Export.hh"
#include "gz/sensors/CameraSensor.hh"
#include "gz/sensors/Export.hh"
#include "gz/sensors/Sensor.hh"

namespace gz
{
  namespace sensors
  {
    // Inline bracket to help doxygen filtering.
    inline namespace GZ_SENSORS_VERSION_NAMESPACE {
    // forward declarations
    class WideAngleCameraSensorPrivate;

    /// \brief Wide Angle camera sensor class.
    ///
    /// This class creates wide angle camera image from a Gazebo Rendering
    /// scene. The scene must be created in advance and given to Manager::Init()
    /// It offers both a gz-transport interface and a direct C++ API
    /// to access the image data. The API works by setting a callback to be
    /// called with image data.
    class GZ_SENSORS_WIDE_ANGLE_CAMERA_VISIBLE WideAngleCameraSensor
      : public CameraSensor
    {
      /// \brief constructor
      public: WideAngleCameraSensor();

      /// \brief destructor
      public: virtual ~WideAngleCameraSensor();

      /// \brief Load the sensor based on data from an sdf::Sensor object.
      /// \param[in] _sdf SDF Sensor parameters.
      /// \return true if loading was successful
      public: virtual bool Load(const sdf::Sensor &_sdf) override;

      /// \brief Load the sensor with SDF parameters.
      /// \param[in] _sdf SDF Sensor parameters.
      /// \return true if loading was successful
      public: virtual bool Load(sdf::ElementPtr _sdf) override;

      /// \brief Initialize values in the sensor
      /// \return True on success
      public: virtual bool Init() override;

      /// \brief Force the sensor to generate data
      /// \param[in] _now The current time
      /// \return true if the update was successfull
      public: virtual bool Update(
        const std::chrono::steady_clock::duration &_now) override;

      /// \brief Get a pointer to the rendering wide angle camera
      /// \return Wide angle camera
      public: virtual rendering::WideAngleCameraPtr WideAngleCamera() const;

      /// \brief Image data callback used to get the data from the sensor
      /// \param[in] _data pointer to the data from the sensor
      /// \param[in] _width width of the image
      /// \param[in] _height height of the image
      /// \param[in] _channel bytes used for the image data
      /// \param[in] _format string with the format
      public: void OnNewWideAngleFrame(const unsigned char *_data,
                    unsigned int _width, unsigned int _height,
                    unsigned int _channels,
                    const std::string &_format);

      /// \brief Set a callback to be called when image frame data is
      /// generated.
      /// \param[in] _callback This callback will be called every time the
      /// camera produces image data. The Update function will be blocked
      /// while the callbacks are executed.
      /// \remark Do not block inside of the callback.
      /// \return A connection pointer that must remain in scope. When the
      /// connection pointer falls out of scope, the connection is broken.
      public: common::ConnectionPtr ConnectImageCallback(
                  std::function<void(const msgs::Image &)> _callback);

      /// \brief Set the rendering scene.
      /// \param[in] _scene Pointer to the scene
      public: virtual void SetScene(
                  gz::rendering::ScenePtr _scene) override;

      /// \brief Get image width.
      /// \return width of the image
      public: virtual unsigned int ImageWidth() const override;

      /// \brief Get image height.
      /// \return height of the image
      public: virtual unsigned int ImageHeight() const override;

      // Documentation inherited.
      public: rendering::CameraPtr RenderingCamera() const override;

      // Documentation inherited.
      public: virtual bool HasConnections() const override;

      /// \brief Create a camera in a scene
      /// \return True on success.
      private: bool CreateCamera();

      GZ_UTILS_WARN_IGNORE__DLL_INTERFACE_MISSING
      /// \brief Data pointer for private data
      /// \internal
      private: std::unique_ptr<WideAngleCameraSensorPrivate> dataPtr;
      GZ_UTILS_WARN_RESUME__DLL_INTERFACE_MISSING
    };
    }
  }
}

#endif
